from typing import Optional

import yaml

from opennem.core.loader import load_data


class SettingsNotFound(Exception):
    pass


def load_logging_config(
    filename: str = "logging.yml", fail_silent: bool = True
) -> Optional[dict]:
    """"""

    settings_file_content = load_data(filename, from_settings=True)

    if not settings_file_content:
        if fail_silent:
            return None

        raise SettingsNotFound(
            "Not a valid logging settings file: {}".format(filename)
        )

    config_data = yaml.safe_load(settings_file_content)

    return config_data


def get_logging_config():
    return LOGGING_CONFIG


LOGGING_CONFIG = load_logging_config()



import csv
from typing import Dict, List, Optional

from pydantic.main import BaseModel


class AEMOTableSchema(BaseModel):
    name: str
    namespace: Optional[str]
    fieldnames: List[str]
    records: List[Dict]


class AEMOTableSet(BaseModel):
    tables: List[AEMOTableSchema]


def parse_aemo_csv(content: str) -> AEMOTableSet:
    content_split = content.splitlines()

    datacsv = csv.reader(content_split)

    for row in datacsv:
        if not row or type(row) is not list or len(row) < 1:
            continue

        record_type = row[0]

        if record_type == "C":
            # @TODO csv meta stored in table
            if table["name"] is not None:
                table_name = table["name"]

                if table_name in item["tables"]:
                    item["tables"][table_name]["records"] += table["records"]
                else:
                    item["tables"][table_name] = table

        elif record_type == "I":
            if table["name"] is not None:
                table_name = table["name"]

                if table_name in item["tables"]:
                    item["tables"][table_name]["records"] += table["records"]
                else:
                    item["tables"][table_name] = table

            table = {}
            table["name"] = "{}_{}".format(row[1], row[2])
            table["fields"] = fields = row[4:]
            table["records"] = []

        elif record_type == "D":
            values = row[4:]
            record = dict(zip(table["fields"], values))

            table["records"].append(record)





import logging
from collections import OrderedDict
from datetime import datetime, timezone
from textwrap import dedent
from typing import List, Optional, Union

from pytz import timezone as pytz_timezone
from sqlalchemy.orm import Session

from opennem.db import get_database_engine
from opennem.db.models.opennem import FacilityScada, Station
from opennem.schema.network import NetworkSchema
from opennem.schema.time import TimeInterval, TimePeriod
from opennem.schema.units import UnitDefinition
from opennem.utils.cache import cache_scada_result
from opennem.utils.time import human_to_timedelta
from opennem.utils.timezone import is_aware, make_aware
from opennem.utils.version import get_version

from .schema import (
    DataQueryResult,
    OpennemData,
    OpennemDataHistory,
    OpennemDataSet,
    ScadaDateRange,
)

logger = logging.getLogger(__name__)


def stats_factory(
    stats: List[DataQueryResult],
    interval: TimeInterval,
    units: UnitDefinition,
    period: Optional[TimePeriod] = None,
    network: Optional[NetworkSchema] = None,
    timezone: Optional[Union[timezone, pytz_timezone]] = None,
    code: Optional[str] = None,
    region: Optional[str] = None,
    fueltech_group: Optional[bool] = False,
    group_field: Optional[str] = None,
    data_id: Optional[str] = None,
) -> Optional[OpennemDataSet]:
    """
    Takes a list of data query results and returns OpennemDataSets

    @TODO optional groupby field
    @TODO override timezone
    """

    if network:
        timezone = network.get_timezone()

    dates = [s.interval for s in stats]

    if not dates:
        return None

    start = min(dates)
    end = max(dates)

    # should probably make sure these are the same TZ
    if timezone and not is_aware(start):
        start = make_aware(min(dates), timezone)

    if timezone and not is_aware(end):
        end = make_aware(max(dates), timezone)

    # free
    dates = []

    group_codes = list(set([i.group_by for i in stats if i.group_by]))

    stats_grouped = []

    for group_code in group_codes:

        data_grouped = dict()

        for stat in stats:
            if stat.interval not in data_grouped:
                data_grouped[stat.interval] = None

            if stat.group_by == group_code:
                data_grouped[stat.interval] = stat.result

        data_sorted = OrderedDict(sorted(data_grouped.items()))

        data_value = list(data_sorted.values())

        if len([i for i in data_value if i]) == 0:
            continue

        history = OpennemDataHistory(
            start=start,
            last=end,
            interval=interval.interval_human,
            data=data_value,
        )

        data = OpennemData(
            data_type=units.unit_type,
            units=units.unit,
            code=group_code,
            interval=interval,
            period=period,
            history=history,
        )

        if network:
            data.network = network.code

        # *sigh* - not the most flexible model
        # @TODO fix this schema and make it more flexible
        if fueltech_group:
            data.fuel_tech = group_code
            data.id = ".".join(
                [
                    network.code.lower(),
                    "fuel_tech",
                    group_code.lower(),
                    units.unit_type,
                ]
            )
            # @TODO make this an alias
            data.type = units.unit_type

        if group_field:
            group_fields = []

            # setattr(data, group_field, group_code)

            if network:
                group_fields.append(network.code.lower())

            group_fields = group_fields + [
                group_field,
                group_code,
                units.unit_type,
            ]

            data.id = ".".join([f for f in group_fields if f])
            data.type = units.unit_type

        if data_id:
            data.id = data_id

        if not data.id and region:
            _id_list = []

            if network:
                _id_list.append(network.code.lower())

            if region:
                _id_list.append(region.lower())

            if units and units.name_alias:
                _id_list.append(units.name_alias)

            data.id = ".".join(_id_list)

        if region:
            data.region = region

        stats_grouped.append(data)

    dt_now = datetime.now()

    if network:
        dt_now = dt_now.astimezone(network.get_timezone())

    stat_set = OpennemDataSet(
        type=units.unit_type,
        data=stats_grouped,
        created_at=dt_now,
        code=code,
        version=get_version(),
    )

    if network:
        stat_set.network = network.code

    if region:
        stat_set.region = region

    return stat_set


def networks_to_in(networks: List[NetworkSchema]):
    codes = ["'{}'".format(n.code) for n in networks]

    return ", ".join(codes)


@cache_scada_result
def get_scada_range(
    network: Optional[NetworkSchema] = None,
    networks: Optional[List[NetworkSchema]] = None,
    network_region: Optional[str] = None,
) -> Optional[ScadaDateRange]:
    engine = get_database_engine()

    __query = """
        select
            min(fs.trading_interval)::timestamp AT TIME ZONE '{timezone}',
            max(fs.trading_interval)::timestamp AT TIME ZONE '{timezone}'
        from facility_scada fs
        where
            {network_query}
            {network_region_query}
            facility_code not like 'ROOFTOP_%%'
    """

    network_query = ""

    if network:
        network_query = f"fs.network_id = '{network.code}' and"

    if networks:
        network_query = "fs.network_id IN ({}) and ".format(
            networks_to_in(networks)
        )

    network_region_query = ""

    if network_region:
        # @TODO support network regions in scada_range
        # technically we don't need this atm
        pass
        # network_region_query = (
        # f"f.network_region = '{network_region.code}' and"
        # )

    scada_range_query = dedent(
        __query.format(
            network_query=network_query,
            network_region_query=network_region_query,
            timezone="UTC",
        )
    )

    with engine.connect() as c:
        logger.debug(scada_range_query)
        scada_range_result = list(c.execute(scada_range_query))

        if len(scada_range_result) < 1:
            raise Exception("No scada range result")

        scada_min = scada_range_result[0][0]
        scada_max = scada_range_result[0][1]

    if not scada_min or not scada_max:
        return None

    scada_range = ScadaDateRange(
        start=scada_min, end=scada_max, network=network
    )

    return scada_range


def station_attach_stats(station: Station, session: Session) -> Station:
    # @TODO update for new queries
    since = datetime.now() - human_to_timedelta("7d")

    facility_codes = list(set([f.code for f in station.facilities]))

    stats = (
        session.query(FacilityScada)
        .filter(FacilityScada.facility_code.in_(facility_codes))
        .filter(FacilityScada.trading_interval >= since)
        .order_by(FacilityScada.facility_code)
        .order_by(FacilityScada.trading_interval)
        .all()
    )

    for facility in station.facilities:
        facility_power = list(
            filter(lambda s: s.facility_code == facility.code, stats)
        )

        # facility.scada_power = stats_factory(facility_power)

    return station



"""
    settings files - read settings from env


    will read in order:

        * .env
        * .env.{environment}
        * system env
        * pydantic settings

"""


import logging.config
import os
from pathlib import Path

from dotenv import load_dotenv

from opennem.settings.log import LOGGING_CONFIG
from opennem.settings.utils import load_env_file

from .schema import OpennemSettings

if Path(".env").is_file():
    load_dotenv()

MODULE_DIR = os.path.dirname(__file__)

ENV = os.getenv("ENV", default="development")

_env_file = load_env_file(ENV)

# @TODO add logging
if _env_file:
    load_dotenv(dotenv_path=_env_file, override=True)

# @NOTE don't use pydantics env file support since it doesn't support multiple
settings: OpennemSettings = OpennemSettings()


# setup logging

if LOGGING_CONFIG:
    if settings.env == "development":
        LOGGING_CONFIG["root"]["level"] = "DEBUG"
        LOGGING_CONFIG["loggers"]["opennem"]["level"] = "DEBUG"

    logging.config.dictConfig(LOGGING_CONFIG)



from enum import Enum
from pathlib import Path
from typing import Optional

import tomlkit
from packaging.version import LegacyVersion, parse


def _get_project_meta():
    pyproject = Path(__file__).parent / "../../pyproject.toml"

    if not pyproject.is_file():
        return None

    with pyproject.open() as pp_file:
        file_contents = pp_file.read()

    return tomlkit.parse(file_contents)["tool"]["poetry"]


pkg_meta = _get_project_meta()
version = None

if pkg_meta:
    project = str(pkg_meta["name"])
    copyright = "2020, OpenNEM"
    version = str(pkg_meta["version"])

release = version


class VersionPart(Enum):
    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"


def get_version(version_part: Optional[VersionPart] = None) -> str:
    if not version:
        return ""

    version_parsed = parse(version)

    if not version_part:
        return version

    if isinstance(version_parsed, LegacyVersion):
        return version

    if version_part == VersionPart.MAJOR:
        return str(version_parsed.major)

    if version_part == VersionPart.MINOR:
        return str(version_parsed.minor)

    if version_part == VersionPart.PATCH:
        return str(version_parsed.minor)

    return version



import sentry_sdk
from sentry_sdk.integrations.redis import RedisIntegration
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration

from opennem.settings import settings

if settings.debug is False:
    sentry_sdk.init(
        settings.sentry_url,
        traces_sample_rate=1.0,
        environment=settings.env,
        integrations=[RedisIntegration(), SqlalchemyIntegration()],
    )



# pylint: disable=no-name-in-module
# pylint: disable=no-self-argument
# pylint: disable=no-member
from huey import RedisHuey, crontab

from opennem.api.export.map import PriorityType
from opennem.api.export.tasks import (
    export_energy,
    export_metadata,
    export_power,
)
from opennem.exporter.geojson import export_facility_geojson
from opennem.monitors.aemo_intervals import get_wem_interval_delay
from opennem.monitors.opennem_metadata import check_metadata_status
from opennem.settings import settings

redis_host = None

if settings.cache_url:
    redis_host = settings.cache_url.host

huey = RedisHuey("opennem.exporter", host=redis_host)
# export tasks


@huey.periodic_task(crontab(minute="*/2"))
@huey.lock_task("schedule_live_tasks")
def schedule_live_tasks():
    export_power(PriorityType.live)


@huey.periodic_task(crontab(hour="*/1"))
@huey.lock_task("schedule_hourly_tasks")
def schedule_hourly_tasks():
    export_energy(PriorityType.daily, latest=True)


@huey.periodic_task(crontab(hour="*/6"))
@huey.lock_task("schedule_daily_tasks")
def schedule_daily_tasks():
    export_energy(PriorityType.daily)


@huey.periodic_task(crontab(hour="*/6"))
@huey.lock_task("schedule_wem_export_all")
def schedule_wem_export_all():
    export_energy(PriorityType.monthly)


@huey.periodic_task(crontab(minute="*/30"))
@huey.lock_task("schedule_export_geojson")
def schedule_export_geojson():
    export_facility_geojson()


@huey.periodic_task(crontab(minute="*/30"))
@huey.lock_task("schedule_export_metadata")
def schedule_export_metadata():
    export_metadata()


# monitoring tasks
@huey.periodic_task(crontab(minute="*/1"))
@huey.lock_task("monitor_wem_interval")
def monitor_wem_interval():
    get_wem_interval_delay()


@huey.periodic_task(crontab(hour="*/1"))
@huey.lock_task("monitor_metadata_status")
def monitor_metadata_status():
    check_metadata_status()



import logging
from datetime import datetime
from typing import Optional

from opennem.api.export.controllers import (
    energy_fueltech_daily,
    power_week,
    weather_daily,
)
from opennem.api.export.map import PriorityType, StatType, get_export_map
from opennem.api.export.utils import write_output
from opennem.settings import settings

YEAR_MIN = 2010

logger = logging.getLogger(__name__)

export_map = get_export_map()


def export_power(priority: Optional[PriorityType] = None):
    power_stat_sets = export_map.get_by_stat_type(StatType.power, priority)

    for power_stat in power_stat_sets:

        stat_set = power_week(
            date_range=power_stat.date_range,
            period=power_stat.period,
            network_code=power_stat.network.code,
            network_region_code=power_stat.network_region,
            networks_query=power_stat.networks,
        )

        if not stat_set:
            continue

        if power_stat.bom_station:
            weather_set = weather_daily(
                station_code=power_stat.bom_station,
                network_code=power_stat.network.code,
                include_min_max=False,
                period_human="7d",
                unit_name="temperature",
            )

            stat_set.append_set(weather_set)

        # print(power_stat.path)
        write_output(power_stat.path, stat_set)


def export_energy(
    priority: Optional[PriorityType] = None, latest: bool = False
):
    energy_stat_sets = export_map.get_by_stat_type(StatType.energy, priority)

    CURRENT_YEAR = datetime.now().year

    for energy_stat in energy_stat_sets:
        if energy_stat.year:

            if latest and energy_stat.year != CURRENT_YEAR:
                continue

            stat_set = energy_fueltech_daily(
                year=energy_stat.year,
                network=energy_stat.network,
                networks_query=energy_stat.networks,
                network_region_code=energy_stat.network_region,
            )

            if not stat_set:
                continue

            if energy_stat.bom_station:
                weather_stats = weather_daily(
                    station_code=energy_stat.bom_station,
                    year=energy_stat.year,
                    network_code=energy_stat.network.code,
                )
                stat_set.append_set(weather_stats)

            write_output(energy_stat.path, stat_set)

        elif energy_stat.period and energy_stat.period.period_human == "all":
            stat_set = energy_fueltech_daily(
                interval_size="1M",
                network=energy_stat.network,
                networks_query=energy_stat.networks,
                network_region_code=energy_stat.network_region,
            )

            if not stat_set:
                continue

            if energy_stat.bom_station:
                weather_stats = weather_daily(
                    station_code=energy_stat.bom_station,
                    year=energy_stat.year,
                    network_code=energy_stat.network.code,
                )
                stat_set.append_set(weather_stats)

            write_output(energy_stat.path, stat_set)


def export_metadata():

    _export_map_out = export_map

    # this is a hack because pydantic doesn't
    # serialize properties
    for r in _export_map_out.resources:
        r.file_path = r.path

    write_output("metadata.json", _export_map_out)


if __name__ == "__main__":
    if settings.env in ["development", "staging"]:
        # export_power(priority=PriorityType.live)
        export_power(priority=PriorityType.history)
        # export_energy(latest=True)
        # export_metadata()
        # wem_export_power()
        # wem_export_daily(limit=1)
        # wem_export_monthly()
    else:
        export_power()
        export_energy()
